from typing import TYPE_CHECKING, Generic, Optional, Tuple, TypeVar

if TYPE_CHECKING:
    Self = TypeVar("Self")

    T_start = TypeVar("T_start")
    T_stop = TypeVar("T_stop")
    T_step = TypeVar("T_step")

    class Slice(slice, Generic[T_start, T_stop, T_step]):
        start: T_start
        stop: T_stop
        step: T_step

    MessageIndex = Tuple[int, Optional[int]]



class InvalidEventTypeDefinition(Exception):
    "不合法的事件类型定义."


class InvalidVerifyKey(Exception):
    "无效的 verifyKey 或其配置."


class AccountNotFound(Exception):
    "未能使用所配置的账号激活 sessionKey, 请检查 mirai_session 配置."


class InvalidSession(Exception):
    "无效的 sessionKey, 请重新获取."


class UnVerifiedSession(Exception):
    "尚未验证/绑定的 session."


class UnknownTarget(Exception):
    "对象位置未知, 不存在或不可及."


class AccountMuted(Exception):
    "账号在对象所在聊天区域被封禁."


class MessageTooLong(Exception):
    "消息过长, 尝试分段发送或报告问题."


class InvalidArgument(Exception):
    "操作参数不合法, 请报告问题."


class NotSupportedAction(Exception):
    "该版本不支持本接口."


class DeprecatedImpl(Exception):
    "该接口已弃用."


class MissingNecessaryArgument(Exception):
    "应在所提到的参数之中至少传入/使用一个"


class ConflictItem(Exception):
    "项冲突/其中一项被重复定义"


class UnknownError(Exception):
    "其他错误"



from graia.broadcast import Dispatchable
from pydantic import validator

from ..dispatcher import ApplicationDispatcher
from ..exception import InvalidEventTypeDefinition
from ..model import AriadneBaseModel


class MiraiEvent(Dispatchable, AriadneBaseModel):
    type: str

    @validator("type", allow_reuse=True)
    def validate_event_type(cls, v):
        if cls.type != v:
            raise InvalidEventTypeDefinition(
                "{0}'s type must be '{1}', not '{2}'".format(cls.__name__, cls.type, v)
            )
        return v

    class Config:
        extra = "ignore"

    Dispatcher = ApplicationDispatcher



import abc
import copy
import re
from argparse import Action
from contextvars import ContextVar
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)

from .util import gen_flags_repr

if TYPE_CHECKING:
    from ...typing import Self
    from ..chain import MessageChain
    from ..element import Element


@dataclass(init=True, eq=True, repr=True)
class ParamPattern:
    longs: List[str]
    short: Optional[str] = None
    default: Any = None
    help_message: Optional[str] = None


@dataclass(init=True, eq=True, repr=True)
class SwitchParameter(ParamPattern):
    default: bool = False
    auto_reverse: bool = False


@dataclass(init=True, eq=True, repr=True)
class BoxParameter(ParamPattern):
    "可以被指定传入消息的参数, 但只有一个."


class Match(abc.ABC):
    pattern: str
    optional: bool
    matched: Optional[bool]
    result: Optional["MessageChain"]

    def __init__(self, pattern, optional: bool = False) -> None:
        self.pattern = pattern
        self.optional = optional
        self.result = None
        self.matched = None
        if self.__class__ == Match:
            raise ValueError("You can't instantiate Match class directly!")

    def __repr__(self) -> str:
        values = {
            k: v
            for k, v in self.__dict__.items()
            if not k.startswith("_") and not callable(v)
        }
        return f"<{self.__class__.__qualname__} {values})>"

    def clone(self, result: "MessageChain", matched: bool) -> "Self":
        new_instance = copy.copy(self)
        new_instance.result = result
        new_instance.matched = matched
        return new_instance

    def gen_regex(self) -> str:
        ...


class RegexMatch(Match):
    "基础的正则表达式匹配器"
    flags: re.RegexFlag
    flags_repr: str
    regex_match: Optional[re.Match]
    preserve_space: bool

    def __init__(
        self,
        pattern: str,
        *,
        optional: bool = False,
        flags: re.RegexFlag = re.RegexFlag(0),
        preserve_space: bool = True,
    ) -> None:
        super().__init__(pattern, optional)
        self.flags = flags
        self.flags_repr = gen_flags_repr(self.flags)
        self.regex_match = None
        self.preserve_space = preserve_space

    def gen_regex(self) -> str:
        return (
            f"({f'?{self.flags_repr}:' if self.flags_repr else ''}{self.pattern})"
            f"{'?' if self.optional else ''}{'( )?' if self.preserve_space else ''}"
        )

    def clone(
        self, result: "MessageChain", matched: bool, re_match: Optional[re.Match] = None
    ) -> "Self":
        new_instance: RegexMatch = super().clone(result, matched)
        new_instance.regex_match = re_match
        return new_instance


class WildcardMatch(Match):
    """泛匹配."""

    preserve_space: bool

    def __init__(
        self, greed: bool = True, optional: bool = False, preserve_space: bool = True
    ) -> None:
        super().__init__(f".*", optional=optional)
        self.greed = greed
        self.preserve_space = preserve_space

    def gen_regex(self) -> str:
        return f"({self.pattern}{'?' if self.greed else ''}){'?' if self.optional else ''}{'( )?' if self.preserve_space else ''}"


class FullMatch(Match):
    """全匹配."""

    def __init__(
        self, pattern: str, *, optional: bool = False, preserve_space: bool = True
    ) -> None:
        super().__init__(pattern, optional)
        self.preserve_space = preserve_space

    def gen_regex(self) -> str:
        return f"({re.escape(self.pattern)}){'?' if self.optional else ''}{'( )?' if self.preserve_space else ''}"


class ElementMatch(Match):
    """元素类型匹配."""

    pattern: Type["Element"]
    result: "Element"

    def __init__(
        self,
        pattern: Type["Element"],
        optional: bool = False,
        preserve_space: bool = True,
    ) -> None:
        super().__init__(pattern, optional=optional)
        self.preserve_space = preserve_space

    def gen_regex(self) -> str:
        return (
            f"(\x02\\d+_{self.pattern.__fields__['type'].default}\x03){'?' if self.optional else ''}"
            f"{'( )?' if self.preserve_space else ''}"
        )


T_const = TypeVar("T_const")
T_default = TypeVar("T_default")


class ArgumentMatch(Match):
    """参数匹配."""

    pattern: Sequence[str]
    name: str
    nargs: Union[str, int]
    action: Union[str, Type[Action]]
    const: Optional[T_const]
    default: Optional[T_default]
    regex: Optional[str]
    result: Union["MessageChain", Any]
    add_arg_data: Dict[str, Any]
    elem_mapping_ctx: ContextVar["MessageChain"] = ContextVar("elem_mapping_ctx")

    def __init__(
        self,
        *pattern: str,
        optional: bool = True,
        const: Optional[T_const] = ...,
        default: Optional[T_default] = ...,
        nargs: Union[str, int] = ...,
        action: Union[str, Type[Action]] = ...,
        regex: Optional[str] = None,
    ) -> None:
        if not pattern:
            raise ValueError("Expected at least 1 pattern!")
        super().__init__(pattern, optional)
        self.name = pattern[0].lstrip("-").replace("-", "_")
        self.nargs = nargs
        self.action = action
        self.const = const
        self.default = default
        self.regex = re.compile(regex) if regex else None
        data: Dict[str, Any] = {}
        if action is not ...:
            data["action"] = action
        if nargs is not ...:
            data["nargs"] = nargs
        if const is not ...:
            data["const"] = const
        if default is not ...:
            data["default"] = default
        if pattern[0].startswith("-"):
            data["required"] = not optional
        self.add_arg_data = data



"""本模块提供 Ariadne 内部使用的小工具, 以及方便的 `async_exec` 模块.
"""

# Utility Layout
import functools
import sys
import traceback
from asyncio.events import AbstractEventLoop
from typing import Callable, ContextManager, Dict, Generator, List, Type, TypeVar, Union

from graia.broadcast import Broadcast
from graia.broadcast.entities.decorator import Decorator
from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.entities.event import Dispatchable
from graia.broadcast.entities.listener import Listener
from graia.broadcast.entities.namespace import Namespace
from graia.broadcast.interfaces.dispatcher import DispatcherInterface
from graia.broadcast.typing import T_Dispatcher
from loguru import logger
from typing_extensions import ParamSpec

from ..context import enter_context

# Import layout
from . import async_exec
from .async_exec import ParallelExecutor, cpu_bound, io_bound

P = ParamSpec("P")
R = TypeVar("R")

from ..exception import (
    AccountMuted,
    AccountNotFound,
    InvalidArgument,
    InvalidSession,
    InvalidVerifyKey,
    MessageTooLong,
    UnknownError,
    UnknownTarget,
    UnVerifiedSession,
)

code_exceptions_mapping: Dict[int, Type[Exception]] = {
    1: InvalidVerifyKey,
    2: AccountNotFound,
    3: InvalidSession,
    4: UnVerifiedSession,
    5: UnknownTarget,
    6: FileNotFoundError,
    10: PermissionError,
    20: AccountMuted,
    30: MessageTooLong,
    400: InvalidArgument,
}


def validate_response(code: Union[dict, int]):
    origin = code
    if isinstance(code, dict):
        code = code.get("code")
    else:
        code = code
    if not isinstance(code, int) or code == 200 or code == 0:
        return
    exc_cls = code_exceptions_mapping.get(code)
    if exc_cls:
        raise exc_cls(exc_cls.__doc__)
    else:
        raise UnknownError(f"{origin}")


def loguru_excepthook(cls, val, tb, *_, **__):
    logger.opt(exception=(cls, val, tb)).error(f"Exception:")


def loguru_async_handler(loop: AbstractEventLoop, ctx: dict):
    logger.opt(exception=(Exception, ctx["message"], ctx["source_traceback"])).error(
        f"Exception:"
    )


def inject_loguru_traceback(loop: AbstractEventLoop):
    """使用 loguru 模块 替换默认的 traceback.print_exception 与 sys.excepthook"""
    traceback.print_exception = loguru_excepthook
    sys.excepthook = loguru_excepthook
    loop.set_exception_handler(loguru_async_handler)


def inject_bypass_listener(broadcast: Broadcast):
    """注入 BypassListener 以享受子事件解析.

    Args:
        broadcast (Broadcast): 外部事件系统, 提供了 event_class_generator 方法以生成子事件.
    """

    class BypassListener(Listener):
        def __init__(
            self,
            callable: Callable,
            namespace: Namespace,
            listening_events: List[Type[Dispatchable]],
            inline_dispatchers: List[T_Dispatcher] = None,
            decorators: List[Decorator] = None,
            priority: int = 16,
        ) -> None:
            events = []
            for event in listening_events:
                events.append(event)
                events.extend(broadcast.event_class_generator(event))
            super().__init__(
                callable,
                namespace,
                events,
                inline_dispatchers=inline_dispatchers,
                decorators=decorators,
                priority=priority,
            )

    @broadcast.dispatcher_interface.inject_global_raw
    async def _(interface: DispatcherInterface):
        if isinstance(interface.event, interface.annotation):
            return interface.event
        elif (
            hasattr(interface.annotation, "__origin__")
            and interface.annotation.__origin__ is DispatcherInterface
        ):
            return interface

    import graia.broadcast.entities.listener

    graia.broadcast.entities.listener.Listener = BypassListener
    graia.broadcast.Listener = BypassListener
    try:  # Override saya listener
        import graia.saya.builtins.broadcast.schema

        graia.saya.builtins.broadcast.schema.Listener = BypassListener
    except ImportError:  # Saya not installed, pass.
        pass


class ApplicationMiddlewareDispatcher(BaseDispatcher):
    always = True
    context: ContextManager

    def __init__(self, app) -> None:
        self.app = app

    def beforeExecution(self, interface: "DispatcherInterface"):
        self.context = enter_context(self.app, interface.event)
        self.context.__enter__()

    def afterExecution(self, interface: "DispatcherInterface", exception, tb):
        self.context.__exit__(exception.__class__ if exception else None, exception, tb)

    async def catch(self, interface: "DispatcherInterface"):
        from ..app import Ariadne

        if interface.annotation is Ariadne:
            return self.app


def app_ctx_manager(func: Callable[P, R]) -> Callable[P, R]:
    @functools.wraps(func)
    async def wrapper(self, *args: P.args, **kwargs: P.kwargs):
        with enter_context(app=self):
            return await func(self, *args, **kwargs)

    return wrapper


T = TypeVar("T")


def gen_subclass(cls: Type[T]) -> Generator[Type[T], None, None]:
    yield cls
    for sub_cls in cls.__subclasses__():
        yield from gen_subclass(sub_cls)


def wrap_bracket(string: str) -> str:
    return string.replace("[", "\\u005b").replace("]", "\\u005d")


T_Callable = TypeVar("T_Callable", bound=Callable)


def deprecated(remove_ver: str) -> Callable:
    def out_wrapper(func: T_Callable) -> T_Callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            logger.warning(f"Deprecated function: {func.__qualname__}")
            logger.warning(f"{func.__qualname__} will be removed in {remove_ver}!")
            return func(*args, **kwargs)

        return wrapper

    return out_wrapper
